/*  GtkExText	
* This is my first GtkWidget
* THIS CODE IS BASED ON Scintilla and GtkText widgets 
* Scintilla is written by Neil Hodgson <neilh@hare.enty.eu 
*  GtkExText is Copyright (C) 1999 by:
*Mikael Hermansson <mikeh@bahnhof.se>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#ifndef GTK_EXTEXT_H__
#define GTK_EXTEXT_H__
#ifdef __cplusplus
extern "C" {
#endif

#include <gtk/gtk.h>
#ifdef WITH_REGEX
#include <sys/types.h>
#include <regex.h>
#endif

#define GTK_TYPE_EXTEXT                  (gtk_extext_get_type ())
#define GTK_EXTEXT(obj)                  (GTK_CHECK_CAST ((obj), GTK_TYPE_EXTEXT, GtkExText))
#define GTK_EXTEXT_CLASS(klass)          (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_EXTEXT, GtkExTextClass))
#define GTK_IS_EXTEXT(obj)               (GTK_CHECK_TYPE ((obj), GTK_TYPE_EXTEXT))
#define GTK_IS_EXTEXT_CLASS(klass)       (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_EXTEXT))
#define GTK_EXTEXT_INDEX(a,b) gtk_extext_get_char_at_pos(a,b)

typedef struct _GtkExText GtkExText;
typedef struct _GtkExTextClass GtkExTextClass;

/* GtkExTextStyle Setup font and bg/fg colors */
/* to use in textdocument */
typedef struct _GtkExTextStyle
{
	GdkFont *font;				/* font */
	GdkColor fg_color;    /* bg_color */
	GdkColor bg_color;   /* fg_color */
}GtkExTextStyle;

/* Line cacher struct */
typedef struct _LineCache
{
	gint len;
	gpointer data;
	/* private */
	struct _LineCache *next;
	struct _LineCache *prev;
} LineCache;

/* Property is used  to set a property mark in the textdoc*/
typedef struct _GtkExTextProperty
{
	guint startpos;				/* startpos from begining of text */ 
	guint endpos;		/* endpos from begining of text */

	gchar *style_key;     /* Style key */
	gpointer data;            /*User data */
	
	/* private user should NOT change this */
	struct _GtkExTextProperty *next;
	struct _GtkExTextProperty *prev;

} GtkExTextProperty;

/* GtkExText UndoBuffert cache */
enum {	
			EXTEXT_UNDO_INSERT,	
			EXTEXT_UNDO_REMOVE
};	

typedef struct _UndoBuffert	{
		gint type;
		gint startpos;
		gint endpos;		
		guchar *buffert;
} UndoBuffert;

struct _GtkExText 
{
	GtkEditable editable;

	GdkWindow *text_area;
	GdkGC *gc;	
	GdkGC *bg_gc;	
	GdkFont *font;
	GtkAdjustment *vadj;
	GtkAdjustment *hadj;
	gboolean freeze;
	
	guint timer;
	guint button;

	gint gap_len;
	gint part1len;

	guchar *text;	
	guchar *part2text;

	/* size is how mush allocated text memory */
	gint size;
	/* length is how mush used text memory */
	gint length;

	/* first visible line */		
	gint scroll_line_start;
	/* number of visible lines  scroll_line_start + scroll_line_count = last visible line*/
	gint scroll_line_count;

	/* line cache coding see source for more info */	
	LineCache * line_start;
	LineCache * line_end;
	LineCache *line_pos;
	gint line_pos_index;
	gint line_cursor_index;
	gint line_count;

	/* this is set if widget is readonly */
	/* yet not supported FIXLATER*/
	gboolean read_only;
	
	/* Undo coding */
	/*need to cleanup undo code */
	GSList *undo_buffert;
	gint undo_max;
	gint undo_count;
	gint undo_last;
	gint undo_flag;

	GtkExTextProperty  *property_start;	
	GtkExTextProperty  *property_end;	
	GtkExTextProperty *property_current;

	/* hash list with font/color styles */
	GHashTable *hash_styles;
};
struct _GtkExTextClass
{
	GtkEditableClass parent_class;

	/* SIGNAL CALLBACKS */
	gint (*property_insert)(GtkExText *text,GtkExTextProperty *prop,gint startpos,gint endpos);
	gint (*property_remove)(GtkExText *text,GtkExTextProperty *prop,gint startpos,gint endpos);
	gint (*undo_changed)(GtkExText *text);	
	/* FIXLATER */
	void (*property_mark)(GtkExText *text);
	gint (*line_insert)(GtkExText *text, gint line_number);
	gint (*line_remove)(GtkExText *text, gint line_number);
};

/* the LineCache callback */
/* You should use this stucture if you want check multiple lines */
/* checkout gtk_extext_line_foreach API for more details */

typedef gboolean		(*LineForEachFunc)	(GtkExText *,gint line, /* the line number */
						gint startpos,							/* startpos in textsegment */
						gint endpos,				/* endpos in textsegment */
						 gpointer line_user_data, /* Lineiuserdata */
			 gpointer	user_data); /* your passed pointer  */
																/*from gtk_extext_foreach */


/* GtkWidget API */

GtkType gtk_extext_get_type();

GtkWidget *gtk_extext_new(GtkAdjustment *hadj,GtkAdjustment *vadj);
void       gtk_extext_insert          (GtkExText       *text,
				     GdkFont       *font,
				     GdkColor      *fore,
				     GdkColor      *back,
				     const char    *chars,
				     gint           length);

void gtk_extext_insert_with_style_and_property(GtkExText *text,
				const char *chars,gint len, gchar *stylekey, gpointer data);

void gtk_extext_set_adjustments(GtkExText *text,GtkAdjustment *hadj,GtkAdjustment *vadj);

/* text API */

gint gtk_extext_get_length(GtkExText *text);

/* Editable class Wrapper API */
#define gtk_extext_backward_delete(a,b) gtk_editable_delete_text(GTK_EDITABLE(a),b-1,b)
#define gtk_extext_get_text(a,b,c)	gtk_editable_get_chars(GTK_EDITABLE(a),b,c);

guchar gtk_extext_get_char_at_pos(GtkExText *text,gint pos);

/* Line and column API */

gint gtk_extext_set_line(GtkExText *text,gint pos);
gint gtk_extext_get_line(GtkExText *text);
void gtk_extext_set_line_data(GtkExText *text,gint line,gpointer userdata);

/* Line get data returns userdata pointer */
gpointer gtk_extext_line_get_data(GtkExText *text,gint line);

/* use line_get_data_and_pos if you want to get startpos/endpos + userdata */
gpointer gtk_extext_line_get_data_and_pos(GtkExText *text,gint line,
			gint *startpos,gint *endpos);

/* You should use this if you want to get multiple lines  */
/* For example when you want to search for text */

gint gtk_extext_line_foreach(GtkExText *text,gint lineindex,LineForEachFunc func,gpointer data);

gint gtk_extext_set_column(GtkExText *text,gint col);
gint gtk_extext_get_column(GtkExText *text);

void gtk_extext_set_position(GtkEditable *editable,gint pos);
gint gtk_extext_get_position(GtkEditable *editable);

/* Undo funcs */

void gtk_extext_undo_set_max(GtkExText *text,gint max);
gboolean gtk_extext_undo_is_empty(GtkExText *text);
gboolean gtk_extext_redo_is_empty(GtkExText *text);
void gtk_extext_undo_clear_all(GtkExText *text);
void gtk_extext_undo(GtkExText *text);
void gtk_extext_redo(GtkExText *text);

/* Style functions */

GtkExTextStyle* gtk_extext_style_insert(GtkExText *text,gchar *stylename,GdkFont *font,GdkColor *fg,GdkColor *bg);
GtkExTextStyle *gtk_extext_style_get(GtkExText *text,gchar *key);
void gtk_extext_style_remove(GtkExText *textclass,gchar *key);

/* Property functions */

void gtk_extext_property_change(GtkExText *text,GtkExTextProperty *prop,gint start,gint end,gboolean replaceold);
GtkExTextProperty * gtk_extext_property_insert(GtkExText *text,gchar*key,gint startpos, gint endpos,gpointer data,gboolean replaceold);
GtkExTextProperty * gtk_extext_property_remove(GtkExText *text,GtkExTextProperty *remove);
GtkExTextProperty *gtk_extext_property_get_at_pos(GtkExText *text,gint pos,GtkExTextProperty *cur);
GtkExTextProperty * gtk_extext_property_move_all(GtkExText *text,gint pos, gint diff,GtkExTextProperty *cur);
GtkExTextProperty * gtk_extext_property_remove_all(GtkExText *text,gint start, gint end,GtkExTextProperty *cur);
GtkExTextProperty * gtk_extext_property_nearest_forward(GtkExText *text,gint pos,GtkExTextProperty *cur);
GtkExTextProperty * gtk_extext_property_nearest_backward(GtkExText *text,gint pos,GtkExTextProperty *cur);
GtkExTextProperty *gtk_extext_property_get_current(GtkExText *text);

void gtk_extext_get_next_word(GtkExText *text,gint *startpos,gint *endpos);
void gtk_extext_get_previous_word(GtkExText *text,gint *startpos,gint *endpos);

void gtk_extext_freeze(GtkExText *text);
void gtk_extext_thaw(GtkExText *text);

GtkType gtk_extext_get_type(void);

#ifdef WITH_REGEX
/* Regex functions */
typedef struct _Match {
  gint from;
  gint to;
} Match;
typedef struct _Regex {
  struct re_pattern_buffer buf;
  struct re_registers reg;
} Regex;
gint gtk_extext_regex_search(GtkExText *text,gint start,Regex *regex,gboolean forward,Match *m);
gint gtk_extext_regex_match(GtkExText *text,gint pos,Regex *regex);
#endif

#ifdef __cplusplus
}
#endif

#endif
