/*  Higlight code	
*  Copyright (C) 1999 by:
*Mikael Hermansson <mikeh@bahnhof.se>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <gtk/gtk.h>
#include "gtkextext.h"
#include "highlight.h"

Regex regstart;
Regex regend;

gchar *table1[]={"ifdef","endif","define","ifndef","include","elif",NULL};
gchar *table2[]={"if","else","for","return","void","int","unsigned","case","do","struct","typedef"
						,"while","long","static","bool","short","char","break","cont","float","const","enum",NULL};
gchar *table3[]={"gint","gboolean","guint","gchar","guchar","gfloat",NULL};

gboolean highlight_compile_regex (const gchar *pattern, Regex *regex)
{
  memset (&regex->buf, 0, sizeof(regex->buf));
  regex->buf.translate = NULL;
  regex->buf.fastmap = g_malloc (256);
  regex->buf.allocated = 0;
  regex->buf.buffer = NULL;
  regex->buf.can_be_null = 0;	/* so we wont allow that in patterns! */
  regex->buf.no_sub = 0;
  if (re_compile_pattern (pattern, strlen (pattern), &regex->buf) == 0) {
    /* success...now try to compile a fastmap */
    if ( re_compile_fastmap (&regex->buf) != 0 ) {
      /* error...no fastmap */
      g_free (regex->buf.fastmap);
      regex->buf.fastmap = NULL;
    }
    return TRUE;
  } else {
    return FALSE;
  }
}

void check_syntax(GtkExText *text,gint i,gint end);

gint get_syntax_end(GtkExText *text,gint start, gint type)
{
	guchar key,keyprev;
	gint length,i;

	keyprev=0;
	i=start;
	length=gtk_extext_get_length(text);
	while(i<length)	{
		key=gtk_extext_get_char_at_pos(text,i);
		switch(type)	{
			case DESCRIPTION_LINE:	
				if(key=='\n')	
					return ++i;				
			break;
			case DESCRIPTION_MULTILINE:
				if(key=='/' && keyprev=='*')
					return ++i;
			break;
			case STRING:
				if(key=='\"' && keyprev!='\\') /* if keyprev ==  its probadly in the string */
					return ++i;	
			break;
		};

		i++;
		keyprev=key;		
	};
	return length;
}

gint cb_removetext(GtkExText *text,GtkExTextProperty *prop,gint start,gint end)
{
	GtkExTextProperty *prop_end;

	gtk_extext_property_move_all(text,end,-(end-start),prop);

	prop_end=gtk_extext_property_nearest_forward(text,end,prop);

	start=prop ? prop->startpos : 0;
	end=prop_end ? prop_end->endpos : text->length;

	gtk_extext_property_remove_all(text,start,end,prop);

	check_syntax(text,start,end);

	return TRUE;
}

gint cb_inserttext(GtkExText *text,GtkExTextProperty *prop,gint start,gint end)
{
	/* THIS IS A FUCKING test and is probadly slow */
	/* we need a better word checker */
 	/* should the widget take care about this ??? */

	/* hmmm no i dont think so */
	/* but we could do a new widget based on gtkextext */
	/* with "insert_word/change_word" signal callbaks ??? */

	GtkExTextProperty *prop_end;
	gint diff;

	diff=end-start;
	prop_end=gtk_extext_property_nearest_forward(text,end,prop);

	start=prop ? prop->startpos : 0;
	end=prop_end ? prop_end->endpos + diff : text->length;

	gtk_extext_property_move_all(text,end,diff,prop);
	gtk_extext_property_remove_all(text,start,end,prop);

	check_syntax(text,start,end);


	return TRUE ;            /* I take care */

}

void init_highlight(GtkExText *text)
{
	GdkColor col1={0,0,0xFFFF,00};
	GdkColor col2={0,0,0,0xFFFF};
	GdkColor col3={0,0xFFFF,0,0};
	GdkColor col4={0,0xFFFF,0x8000,0};
	GdkColor col5={0,0x8000,0x0000,0};


	gtk_extext_style_insert(text,"green",NULL,&col1,NULL);
	gtk_extext_style_insert(text,"blue",NULL,&col2,NULL);
	gtk_extext_style_insert(text,"red",NULL,&col3,NULL);
	gtk_extext_style_insert(text,"brown",NULL,&col4,NULL);
	gtk_extext_style_insert(text,"glibsyntax",NULL,&col5,NULL);

	
	gtk_signal_connect(GTK_OBJECT(text),"property-insert",
							GTK_SIGNAL_FUNC(cb_inserttext),NULL);
	gtk_signal_connect(GTK_OBJECT(text),"property-remove",
							GTK_SIGNAL_FUNC(cb_removetext),NULL);


	highlight_compile_regex("/*",&regstart);
	highlight_compile_regex("*/", &regend);
		
}

void destroy_highlight(GtkExText *text)
{

}


void check_syntax(GtkExText *text,gint start,gint end)
{
/*	
	GtkExTextProperty *prop;
	Match m;
		
	if(gtk_extext_regex_search(text,start,&regstart,TRUE,&m)>=0)	{
		prop=gtk_extext_property_insert(text,"red",
						m.from,m.from+2,(gpointer)DESCRIPTION_MULTILINE,TRUE);
	
		if(prop && gtk_extext_regex_search(text,m.to+1,&regend,TRUE,&m)>=0)	{
			prop->endpos=m.to;					
		}
	}
*/
	gint bufstart,bufend,b,i;
	guchar *buf,keyprev,key;

	i=start;
	keyprev=i ? gtk_extext_get_char_at_pos(text,i - 1) : 0;

	/* make suure it really higlight last worrd */
	if(!isalnum(gtk_extext_get_char_at_pos(text,end)))
		end++;

	while(i<end)	{
		key=gtk_extext_get_char_at_pos(text,i);
		if(keyprev=='/' && key=='\*' )	{
			gint s;
			s=i-1;
			i=get_syntax_end(text,i,DESCRIPTION_MULTILINE);				
			gtk_extext_property_insert(text,"red",
						s,i,(gpointer)DESCRIPTION_MULTILINE,TRUE);

			key=0;		
			i--;
		}
		else if(keyprev=='/' && key=='/' )	{
			gint s;
			s=i-1;
			i=get_syntax_end(text,i+1,DESCRIPTION_LINE);					

			gtk_extext_property_insert(text,"red",
						s,i,(gpointer)DESCRIPTION_LINE,TRUE);
			
			i--;
			key=0;
		}
		else if( key=='\"' && keyprev!='\\') 
		{
			gint s;
			s=i;
			i=get_syntax_end(text,i+1,STRING);					
			gtk_extext_property_insert(text,"brown",
						s,i,(gpointer)STRING,FALSE);

			g_print(" %d %d\n",s,i);

			i--;	
		}
		else if(!isalnum(key)) 
		{	
			gint k;
			bufstart=i-1;
			bufend=0;
			gtk_extext_get_previous_word(text,&bufstart,&bufend);		
			if(bufstart)
				k=gtk_extext_get_char_at_pos(text,bufstart-1);
			else
				k=0x20;

			buf=gtk_extext_get_text(text,bufstart,bufend);
			if((k==0x20 || k=='\n' || k=='\r' || k=='\t' || k=='#' || k=='(' || k==',')  && buf)	
			{
				gint len;
				len=strlen(buf);
				if(bufend>start)
					i=bufend;					
				
				if(k=='#')	{
					for(b=0;table1[b];b++)	
					{
						if(len==strlen(table1[b]) && 
							!strncmp((const char *)buf,(const char *)table1[b],strlen(table1[b])))		{
								gtk_extext_property_insert(text,"green",bufstart-1,bufstart+strlen(table1[b])+1,(gpointer)SYNTAX,TRUE);			
						}
					}
				}
				for(b=0;table2[b];b++)	
				{
					if(len==strlen(table2[b]) && 
								!strncmp((const char *)buf,(const char *)table2[b],strlen(table2[b])))	
					{
						gtk_extext_property_insert(text,"blue",bufstart,bufstart+strlen(table2[b]),(gpointer)SYNTAX,TRUE);			
					}
				}
				for(b=0;table3[b];b++)	
				{
					if(len==strlen(table3[b]) && 
								 !strncmp((const char *)buf,(const char *)table3[b],strlen(table2[b])))	
					{
						gtk_extext_property_insert(text,"glibsyntax",bufstart,bufstart+strlen(table3[b]),(gpointer)SYNTAX,TRUE);			
					}
				}
				g_free(buf);
			}
		}
		keyprev=key;
		i++;
	}


/*	 if no new propertys make sure all propertys after the new chars is uptodate */

	
}

